@@ -1,4 +1,4 @@
-# Copyright 2014 Jeffrey Kegler
+# Copyright 2022 Jeffrey Kegler
 # This file is part of Marpa::R2.  Marpa::R2 is free software: you can
 # redistribute it and/or modify it under the terms of the GNU Lesser
 # General Public License as published by the Free Software Foundation,
@@ -15,12 +15,12 @@
 
 package Marpa::R2::MetaAST;
 
-use 5.010;
+use 5.010001;
 use strict;
 use warnings;
 
 use vars qw($VERSION $STRING_VERSION);
-$VERSION        = '2.086000';
+$VERSION        = '12.000000';
 $STRING_VERSION = $VERSION;
 ## no critic(BuiltinFunctions::ProhibitStringyEval)
 $VERSION = eval $VERSION;
@@ -33,17 +33,12 @@ use English qw( -no_match_vars );
 sub new {
     my ( $class, $p_rules_source ) = @_;
     my $meta_recce = Marpa::R2::Internal::Scanless::meta_recce();
-    eval { $meta_recce->read($p_rules_source) } or
-        Marpa::R2::exception("Parse of BNF/Scanless source failed\n", $EVAL_ERROR);
-    if ( $meta_recce->ambiguity_metric() > 1 ) {
-    my $asf = Marpa::R2::ASF->new( { slr => $meta_recce } );
-    say STDERR 'No ASF' if not defined $asf;
-    my $ambiguities = Marpa::R2::Internal::ASF::ambiguities( $asf );
-    my @ambiguities = grep { defined } @{$ambiguities}[0 .. 1 ];
-        Marpa::R2::exception(
-            "Parse of BNF/Scanless source is ambiguous\n",
-            Marpa::R2::Internal::ASF::ambiguities_show( $asf, \@ambiguities )
-        );
+    eval { $meta_recce->read($p_rules_source) }
+        or Marpa::R2::exception( "Parse of BNF/Scanless source failed\n",
+        $EVAL_ERROR );
+    if ( my $ambiguity_status = $meta_recce->ambiguous() ) {
+        Marpa::R2::exception( "Parse of BNF/Scanless source failed:\n",
+            $ambiguity_status );
     }
     my $value_ref = $meta_recce->value();
     Marpa::R2::exception('Parse of BNF/Scanless source failed')
@@ -184,6 +179,13 @@ sub Marpa::R2::Internal::MetaAST_Nodes::reserved_action_name::name {
     return $self->[2];
 }
 
+sub Marpa::R2::Internal::MetaAST_Nodes::reserved_event_name::name {
+    my ( $self, $parse ) = @_;
+    my $name = $self->[2];
+    $name =~ s/\A : /'/xms;
+    return $name;
+}
+
 sub Marpa::R2::Internal::MetaAST_Nodes::action_name::name {
     my ( $self, $parse ) = @_;
     return $self->[2]->name($parse);
@@ -364,10 +366,16 @@ sub Marpa::R2::Internal::MetaAST_Nodes::group_association::evaluate {
 
 sub Marpa::R2::Internal::MetaAST_Nodes::event_specification::evaluate {
     my ($values) = @_;
-    my $child = $values->[2];
-    return bless { event => $child->name() }, $PROTO_ALTERNATIVE;
+    return bless { event => ( $values->[2]->event() ) }, $PROTO_ALTERNATIVE;
 }
 
+sub Marpa::R2::Internal::MetaAST_Nodes::event_initialization::event {
+    my ($values)         = @_;
+    my $event_name       = $values->[2];
+    my $event_initializer = $values->[3];
+    return [$event_name->name(), $event_initializer->on_or_off()],
+} ## end sub Marpa::R2::Internal::MetaAST_Nodes::event_specification::evaluate
+
 sub Marpa::R2::Internal::MetaAST_Nodes::proper_specification::evaluate {
     my ($values) = @_;
     my $child = $values->[2];
@@ -412,6 +420,19 @@ sub Marpa::R2::Internal::MetaAST_Nodes::before_or_after::value {
     return $_[0]->[2];
 }
 
+sub Marpa::R2::Internal::MetaAST_Nodes::event_initializer::on_or_off
+{
+    # die Data::Dumper::Dumper(\@_);
+    my ($values) = @_;
+    my $is_activated = $values->[2];
+    return 1 if not defined $is_activated;
+    return $is_activated->value();
+}
+
+sub Marpa::R2::Internal::MetaAST_Nodes::on_or_off::value {
+    return $_[0]->[2] eq 'off' ? 0 : 1;
+}
+
 sub Marpa::R2::Internal::MetaAST_Nodes::boolean::value {
     return $_[0]->[2];
 }
@@ -459,6 +480,34 @@ sub Marpa::R2::Internal::MetaAST_Nodes::default_rule::evaluate {
     return undef;
 } ## end sub Marpa::R2::Internal::MetaAST_Nodes::default_rule::evaluate
 
+sub Marpa::R2::Internal::MetaAST_Nodes::discard_default_statement::evaluate {
+    my ( $data, $parse ) = @_;
+    my ( $start, $length, $raw_adverb_list ) = @{$data};
+    local $Marpa::R2::Internal::SUBGRAMMAR = 'G1';
+
+    my $adverb_list = $raw_adverb_list->evaluate($parse);
+    if ( exists $parse->{discard_default_adverbs} ) {
+        my $problem_rule = $parse->substring( $start, $length );
+        Marpa::R2::exception(
+            qq{More than one discard default statement is not allowed\n},
+            qq{  This was the rule that caused the problem:\n},
+            qq{  $problem_rule\n}
+        );
+    } ## end if ( exists $parse->{discard_default_adverbs} )
+    $parse->{discard_default_adverbs} = {};
+    ADVERB: for my $key ( keys %{$adverb_list} ) {
+        my $value = $adverb_list->{$key};
+        if ( $key eq 'event' and defined $value ) {
+            $parse->{discard_default_adverbs}->{$key} = $value;
+            next ADVERB;
+        }
+        Marpa::R2::exception(
+            qq{"$key" adverb not allowed as discard default"});
+    } ## end ADVERB: for my $key ( keys %{$adverb_list} )
+    ## no critic(Subroutines::ProhibitExplicitReturnUndef)
+    return undef;
+} ## end sub Marpa::R2::Internal::MetaAST_Nodes::discard_default_statement::evaluate
+
 sub Marpa::R2::Internal::MetaAST_Nodes::lexeme_default_statement::evaluate {
     my ( $data, $parse ) = @_;
     my ( $start, $length, $raw_adverb_list ) = @{$data};
@@ -580,7 +629,7 @@ sub Marpa::R2::Internal::MetaAST_Nodes::priority_rule::evaluate {
                 and grep { !$_ } @mask )
             {
                 Marpa::R2::exception(
-                    qq{hidden symbols are not allowed in lexical rules (rules LHS was "$lhs")}
+                    qq{hidden symbols are not allowed in lexical rules (rule's LHS was "$lhs")}
                 );
             }
             my %hash_rule =
@@ -627,27 +676,24 @@ sub Marpa::R2::Internal::MetaAST_Nodes::priority_rule::evaluate {
             $action //= $default_adverbs->{action};
             if ( defined $action ) {
                 Marpa::R2::exception(
-                    'actions not allowed in lexical rules (rules LHS was "',
-                    $lhs, '")' )
-            if  ( substr $subgrammar, 0, 1 ) eq 'L';
+                    qq{actions not allowed in lexical rules (rule's LHS was "$lhs")}
+                ) if  ( substr $subgrammar, 0, 1 ) eq 'L';
                 $hash_rule{action} = $action;
             } ## end if ( defined $action )
 
             $rank //= $default_adverbs->{rank};
             if ( defined $rank ) {
                 Marpa::R2::exception(
-                    'ranks not allowed in lexical rules (rules LHS was "',
-                    $lhs, '")' )
-            if  ( substr $subgrammar, 0, 1 ) eq 'L';
+                    qq{ranks not allowed in lexical rules (rule's LHS was "$lhs")}
+                ) if  ( substr $subgrammar, 0, 1 ) eq 'L';
                 $hash_rule{rank} = $rank;
             } ## end if ( defined $rank )
 
             $null_ranking //= $default_adverbs->{null_ranking};
             if ( defined $null_ranking ) {
                 Marpa::R2::exception(
-                    'null-ranking allowed in lexical rules (rules LHS was "',
-                    $lhs, '")' )
-            if  ( substr $subgrammar, 0, 1 ) eq 'L';
+                    qq{null-ranking allowed in lexical rules (rule's LHS was "$lhs")}
+                ) if  ( substr $subgrammar, 0, 1 ) eq 'L';
                 $hash_rule{null_ranking} = $null_ranking;
             } ## end if ( defined $rank )
 
@@ -783,27 +829,24 @@ sub Marpa::R2::Internal::MetaAST_Nodes::priority_rule::evaluate {
         $action //= $default_adverbs->{action};
         if ( defined $action ) {
             Marpa::R2::exception(
-                'actions not allowed in lexical rules (rules LHS was "',
-                $lhs, '")' )
-            if  ( substr $subgrammar, 0, 1 ) eq 'L';
+                qq{actions not allowed in lexical rules (rule's LHS was "$lhs")}
+            ) if  ( substr $subgrammar, 0, 1 ) eq 'L';
             $new_xs_rule{action} = $action;
         } ## end if ( defined $action )
 
         $null_ranking //= $default_adverbs->{null_ranking};
         if ( defined $null_ranking ) {
             Marpa::R2::exception(
-                'null-ranking not allowed in lexical rules (rules LHS was "',
-                $lhs, '")' )
-            if  ( substr $subgrammar, 0, 1 ) eq 'L';
+                qq{null-ranking not allowed in lexical rules (rule's LHS was "$lhs")}
+            ) if  ( substr $subgrammar, 0, 1 ) eq 'L';
             $new_xs_rule{null_ranking} = $null_ranking;
         } ## end if ( defined $rank )
 
         $rank //= $default_adverbs->{rank};
         if ( defined $rank ) {
             Marpa::R2::exception(
-                'ranks not allowed in lexical rules (rules LHS was "',
-                $lhs, '")' )
-            if  ( substr $subgrammar, 0, 1 ) eq 'L';
+                qq{ranks not allowed in lexical rules (rule's LHS was "$lhs")}
+            ) if  ( substr $subgrammar, 0, 1 ) eq 'L';
             $new_xs_rule{rank} = $rank;
         } ## end if ( defined $rank )
 
@@ -820,7 +863,13 @@ sub Marpa::R2::Internal::MetaAST_Nodes::priority_rule::evaluate {
         $parse->bless_hash_rule( \%new_xs_rule, $blessing, $naming, $lhs );
 
         my $next_priority = $priority + 1;
+
+        # This is probably a mis-feature.  It probably should be
+        # $next_priority = $priority if $next_priority >= $priority_count;
+        # However, I probably will not change this, because some apps
+        # may be relying on this behavior.
         $next_priority = 0 if $next_priority >= $priority_count;
+
         my $next_exp = $parse->prioritized_symbol( $lhs, $next_priority);
 
         if ( not scalar @arity ) {
@@ -931,27 +980,24 @@ sub Marpa::R2::Internal::MetaAST_Nodes::empty_rule::evaluate {
     $action //= $default_adverbs->{action};
     if ( defined $action ) {
         Marpa::R2::exception(
-            'actions not allowed in lexical rules (rules LHS was "',
-            $lhs, '")' )
-            if  ( substr $subgrammar, 0, 1 ) eq 'L';
+            qq{actions not allowed in lexical rules (rule's LHS was "$lhs")}
+        ) if  ( substr $subgrammar, 0, 1 ) eq 'L';
         $rule{action} = $action;
     } ## end if ( defined $action )
 
     $null_ranking //= $default_adverbs->{null_ranking};
     if ( defined $null_ranking ) {
         Marpa::R2::exception(
-            'null-ranking not allowed in lexical rules (rules LHS was "',
-            $lhs, '")' )
-            if  ( substr $subgrammar, 0, 1 ) eq 'L';
+            qq{null-ranking not allowed in lexical rules (rule's LHS was "$lhs")}
+        ) if  ( substr $subgrammar, 0, 1 ) eq 'L';
         $rule{null_ranking} = $null_ranking;
     } ## end if ( defined $null_ranking )
 
     $rank //= $default_adverbs->{rank};
     if ( defined $rank ) {
         Marpa::R2::exception(
-            'ranks not allowed in lexical rules (rules LHS was "',
-            $lhs, '")' )
-            if  ( substr $subgrammar, 0, 1 ) eq 'L';
+            qq{ranks not allowed in lexical rules (rule's LHS was "$lhs")}
+        ) if  ( substr $subgrammar, 0, 1 ) eq 'L';
         $rule{rank} = $rank;
     } ## end if ( defined $rank )
 
@@ -960,8 +1006,8 @@ sub Marpa::R2::Internal::MetaAST_Nodes::empty_rule::evaluate {
         and ( substr $subgrammar, 0, 1 ) eq 'L' )
     {
         Marpa::R2::exception(
-            'bless option not allowed in lexical rules (rules LHS was "',
-            $lhs, '")' );
+            qq{bless option not allowed in lexical rules (rule's LHS was "$lhs")}
+        );
     }
     $parse->bless_hash_rule( \%rule, $blessing, $naming, $lhs );
 
@@ -1097,7 +1143,8 @@ sub Marpa::R2::Internal::MetaAST_Nodes::start_rule::evaluate {
 
 sub Marpa::R2::Internal::MetaAST_Nodes::discard_rule::evaluate {
     my ( $values, $parse ) = @_;
-    my ( $start, $length, $symbol ) = @{$values};
+    my ( $start, $length, $symbol, $raw_adverb_list ) = @{$values};
+
     my $lexer_name = $parse->{current_lexer};
     local $Marpa::R2::Internal::SUBGRAMMAR = $lexer_name;
     my $discard_lhs = '[:discard]';
@@ -1108,16 +1155,30 @@ sub Marpa::R2::Internal::MetaAST_Nodes::discard_rule::evaluate {
             description  => qq{Internal LHS for lexer "$lexer_name" discard}
         }
     );
-    my $rhs = $symbol->names($parse);
-    push @{ $parse->{rules}->{$lexer_name} },
-        {
+    my $rhs         = $symbol->names($parse);
+    my $rhs_as_event         = $symbol->event_name($parse);
+    my $adverb_list = $raw_adverb_list->evaluate($parse);
+    my $event;
+    ADVERB: for my $key ( keys %{$adverb_list} ) {
+        my $value = $adverb_list->{$key};
+        if ( $key eq 'event' ) {
+            $event = $value;
+            next ADVERB;
+        }
+        Marpa::R2::exception(
+            qq{"$key" adverb not allowed as discard default"});
+    } ## end ADVERB: for my $key ( keys %{$adverb_list} )
+    my %rule_hash = (
         description => (
             "Discard rule for " . join q{ },
             map { '<' . $_ . '>' } @{$rhs}
         ),
         lhs => $discard_lhs,
-        rhs => $rhs
-        };
+        rhs => $rhs,
+        symbol_as_event => $rhs_as_event
+    );
+    $rule_hash{event} = $event if defined $event;
+    push @{ $parse->{rules}->{$lexer_name} }, \%rule_hash;
     ## no critic(Subroutines::ProhibitExplicitReturnUndef)
     return undef;
 } ## end sub Marpa::R2::Internal::MetaAST_Nodes::discard_rule::evaluate
@@ -1212,27 +1273,24 @@ sub Marpa::R2::Internal::MetaAST_Nodes::quantified_rule::evaluate {
     $action //= $default_adverbs->{action};
     if ( defined $action ) {
         Marpa::R2::exception(
-            'actions not allowed in lexical rules (rules LHS was "',
-            $lhs, '")' )
-            if  ( substr $subgrammar, 0, 1 ) eq 'L';
+            qq{actions not allowed in lexical rules (rule's LHS was "$lhs")}
+        ) if  ( substr $subgrammar, 0, 1 ) eq 'L';
         $sequence_rule{action} = $action;
     } ## end if ( defined $action )
 
     $null_ranking //= $default_adverbs->{null_ranking};
     if ( defined $null_ranking ) {
         Marpa::R2::exception(
-            'null-ranking not allowed in lexical rules (rules LHS was "',
-            $lhs, '")' )
-            if  ( substr $subgrammar, 0, 1 ) eq 'L';
+            qq{null-ranking not allowed in lexical rules (rule's LHS was "$lhs")}
+        ) if  ( substr $subgrammar, 0, 1 ) eq 'L';
         $sequence_rule{null_ranking} = $null_ranking;
     } ## end if ( defined $null_ranking )
 
     $rank //= $default_adverbs->{rank};
     if ( defined $rank ) {
         Marpa::R2::exception(
-            'ranks not allowed in lexical rules (rules LHS was "',
-            $lhs, '")' )
-            if  ( substr $subgrammar, 0, 1 ) eq 'L';
+            qq{ranks not allowed in lexical rules (rule's LHS was "$lhs")}
+        ) if  ( substr $subgrammar, 0, 1 ) eq 'L';
         $sequence_rule{rank} = $rank;
     } ## end if ( defined $rank )
 
@@ -1240,8 +1298,8 @@ sub Marpa::R2::Internal::MetaAST_Nodes::quantified_rule::evaluate {
     if ( defined $blessing and ( substr $subgrammar, 0, 1 ) eq 'L' )
     {
         Marpa::R2::exception(
-            'bless option not allowed in lexical rules (rules LHS was "',
-            $lhs, '")' );
+            qq{bless option not allowed in lexical rules (rule's LHS was "$lhs")}
+        );
     }
     $parse->bless_hash_rule( \%sequence_rule, $blessing, $naming, $lhs_name );
 
@@ -1254,8 +1312,7 @@ sub Marpa::R2::Internal::MetaAST_Nodes::quantified_rule::evaluate {
 sub Marpa::R2::Internal::MetaAST_Nodes::completion_event_declaration::evaluate
 {
     my ( $values, $parse ) = @_;
-    my ( $start, $length, $raw_event_name, $raw_symbol_name ) = @{$values};
-    my $event_name        = $raw_event_name->name();
+    my ( $start, $length, $raw_event, $raw_symbol_name ) = @{$values};
     my $symbol_name       = $raw_symbol_name->name();
     my $completion_events = $parse->{completion_events} //= {};
     if ( defined $completion_events->{$symbol_name} ) {
@@ -1266,15 +1323,14 @@ sub Marpa::R2::Internal::MetaAST_Nodes::completion_event_declaration::evaluate
             "\n",
             "  Problem occurred at line $line, column $column\n";
     } ## end if ( defined $completion_events->{$symbol_name} )
-    $completion_events->{$symbol_name} = $event_name;
+    $completion_events->{$symbol_name} = $raw_event->event();
     ## no critic(Subroutines::ProhibitExplicitReturnUndef)
     return undef;
 } ## end sub Marpa::R2::Internal::MetaAST_Nodes::completion_event_declaration::evaluate
 
 sub Marpa::R2::Internal::MetaAST_Nodes::nulled_event_declaration::evaluate {
     my ( $values, $parse ) = @_;
-    my ( $start, $length, $raw_event_name, $raw_symbol_name ) = @{$values};
-    my $event_name    = $raw_event_name->name();
+    my ( $start, $length, $raw_event, $raw_symbol_name ) = @{$values};
     my $symbol_name   = $raw_symbol_name->name();
     my $nulled_events = $parse->{nulled_events} //= {};
     if ( defined $nulled_events->{$symbol_name} ) {
@@ -1285,7 +1341,7 @@ sub Marpa::R2::Internal::MetaAST_Nodes::nulled_event_declaration::evaluate {
             "\n",
             "  Problem occurred at line $line, column $column\n";
     } ## end if ( defined $nulled_events->{$symbol_name} )
-    $nulled_events->{$symbol_name} = $event_name;
+    $nulled_events->{$symbol_name} = $raw_event->event();
     ## no critic(Subroutines::ProhibitExplicitReturnUndef)
     return undef;
 } ## end sub Marpa::R2::Internal::MetaAST_Nodes::nulled_event_declaration::evaluate
@@ -1293,8 +1349,7 @@ sub Marpa::R2::Internal::MetaAST_Nodes::nulled_event_declaration::evaluate {
 sub Marpa::R2::Internal::MetaAST_Nodes::prediction_event_declaration::evaluate
 {
     my ( $values, $parse ) = @_;
-    my ( $start, $length, $raw_event_name, $raw_symbol_name ) = @{$values};
-    my $event_name        = $raw_event_name->name();
+    my ( $start, $length, $raw_event, $raw_symbol_name ) = @{$values};
     my $symbol_name       = $raw_symbol_name->name();
     my $prediction_events = $parse->{prediction_events} //= {};
     if ( defined $prediction_events->{$symbol_name} ) {
@@ -1305,7 +1360,7 @@ sub Marpa::R2::Internal::MetaAST_Nodes::prediction_event_declaration::evaluate
             "\n",
             "  Problem occurred at line $line, column $column\n";
     } ## end if ( defined $prediction_events->{$symbol_name} )
-    $prediction_events->{$symbol_name} = $event_name;
+    $prediction_events->{$symbol_name} = $raw_event->event();
     ## no critic(Subroutines::ProhibitExplicitReturnUndef)
     return undef;
 } ## end sub Marpa::R2::Internal::MetaAST_Nodes::prediction_event_declaration::evaluate
@@ -1368,6 +1423,18 @@ sub Marpa::R2::Internal::MetaAST_Nodes::single_symbol::name {
     return $symbol->name($parse);
 }
 
+sub Marpa::R2::Internal::MetaAST_Nodes::single_symbol::event_name {
+    my ( $values, $parse ) = @_;
+    my ( undef, undef, $symbol ) = @{$values};
+    return $symbol->event_name($parse);
+}
+
+sub Marpa::R2::Internal::MetaAST_Nodes::single_symbol::literal {
+    my ( $values, $parse ) = @_;
+    my ( $start, $length ) = @{$values};
+    return $parse->substring($start, $length);
+}
+
 sub Marpa::R2::Internal::MetaAST_Nodes::single_symbol::evaluate {
     my ( $values, $parse ) = @_;
     my ( undef, undef, $symbol ) = @{$values};
@@ -1386,6 +1453,11 @@ sub Marpa::R2::Internal::MetaAST_Nodes::symbol::name {
     return $self->[2]->name($parse);
 }
 
+sub Marpa::R2::Internal::MetaAST_Nodes::symbol::event_name {
+    my ( $self, $parse ) = @_;
+    return $self->[2]->name($parse);
+}
+
 sub Marpa::R2::Internal::MetaAST_Nodes::symbol::names {
     my ( $self, $parse ) = @_;
     return $self->[2]->names($parse);
@@ -1425,9 +1497,10 @@ sub Marpa::R2::Internal::MetaAST_Nodes::adverb_list_items::evaluate {
         @adverb_items);
 } ## end sub Marpa::R2::Internal::MetaAST_Nodes::adverb_list::evaluate
 
-sub Marpa::R2::Internal::MetaAST_Nodes::character_class::name {
-    my ( $self, $parse ) = @_;
-    return $self->evaluate($parse)->name($parse);
+sub Marpa::R2::Internal::MetaAST_Nodes::character_class::event_name {
+    my ( $data,  $parse )  = @_;
+    my ( $start, $length ) = @{$data};
+    return $parse->substring( $start, $length );
 }
 
 sub Marpa::R2::Internal::MetaAST_Nodes::character_class::names {
@@ -1435,11 +1508,15 @@ sub Marpa::R2::Internal::MetaAST_Nodes::character_class::names {
     return [ $self->name($parse) ];
 }
 
+sub Marpa::R2::Internal::MetaAST_Nodes::character_class::name {
+    my ( $self, $parse ) = @_;
+    return $self->evaluate($parse)->name($parse);
+}
+
 sub Marpa::R2::Internal::MetaAST_Nodes::character_class::evaluate {
     my ( $values, $parse ) = @_;
     my $character_class = $values->[2];
     my $subgrammar = $Marpa::R2::Internal::SUBGRAMMAR;
-    $DB::single = 1 if not defined $subgrammar;
     if  (( substr $subgrammar, 0, 1 ) eq 'L') {
         return Marpa::R2::Internal::MetaAST::Symbol_List->char_class_to_symbol(
             $parse, $character_class );
